"""

"""

from univariate.missing_value.handler.strategy import MissingValueHandleStrategy, MissingValueHandleStrategyType
from pyspark.sql import DataFrame
import pyspark.sql.functions as F
from univariate.analyzer import AnalysisReport
from univariate.analyzer import DescriptiveStatAnalyzer
from typing import Optional
from pyspark.sql.window import Window


class LinearInterpolation(MissingValueHandleStrategy):
    """ """

    def handle(
        self, missed_ts: DataFrame, time_col_name: str, data_col_name, handled_col_name: str
    ) -> DataFrame:
        """

        :param missed_ts:
        :param time_col_name:
        :param data_col_name:
        :param handled_col_name:
        :return:
        """
        # https://stackoverflow.com/questions/53077639/pyspark-interpolation-of-missing-values-in-pyspark-dataframe-observed
        # create row number over window and a column with row number only for non missing values
        window = Window.orderBy(time_col_name)
        new_df = missed_ts.withColumn('rn', F.row_number().over(window))
        new_df = new_df.withColumn('rn_not_null', F.when(F.col(data_col_name).isNotNull(), F.col('rn')))

        # create relative references to the start value (last value not missing)
        w_start = Window.orderBy(time_col_name).rowsBetween(Window.unboundedPreceding, -1)
        new_df = new_df.withColumn('start_val', F.last(data_col_name, True).over(w_start))
        new_df = new_df.withColumn('start_rn', F.last('rn_not_null', True).over(w_start))

        # create relative references to the end value (first value not missing)
        w_end = Window.orderBy(time_col_name).rowsBetween(0, Window.unboundedFollowing)
        new_df = new_df.withColumn('end_val', F.first(data_col_name, True).over(w_end))
        new_df = new_df.withColumn('end_rn', F.first('rn_not_null', True).over(w_end))

        # create references to gap length and current gap position
        new_df = new_df.withColumn('diff_rn', F.col('end_rn') - F.col('start_rn'))
        new_df = new_df.withColumn('curr_rn', F.col('diff_rn') - (F.col('end_rn') - F.col('rn')))

        # calculate linear interpolation value
        lin_interp_func = (
                    F.round(F.col('start_val') + (F.col('end_val') - F.col('start_val')) / F.col('diff_rn') * F.col('curr_rn'), 3))
        new_df = new_df.withColumn(handled_col_name,
                                   F.when(F.col(data_col_name).isNull(), lin_interp_func).otherwise(F.col(data_col_name)))

        new_df = new_df.drop('rn', 'rn_not_null', 'start_val', 'end_val', 'start_rn', 'end_rn', 'diff_rn', 'curr_rn')
        return new_df



"""

"""

from univariate.strategy.period import PeriodCalcStrategy
from univariate.analyzer import AnalysisReport
from pyspark.sql import DataFrame
from sklearn.cluster import MeanShift, estimate_bandwidth
import numpy as np
import pyspark.sql.functions as F
from pyspark.sql.window import Window
from pyspark import SparkContext, SQLContext
import plotly.express as px
import logging

logger = logging.getLogger()


class ClusteringAndApproximateGCD(PeriodCalcStrategy):
    """ """

    @classmethod
    def calc_period(cls, diff_df: DataFrame, diff_col_name: str) -> AnalysisReport:
        """

        :param diff_df:
        :param diff_col_name:
        :return:
        """
        logger.debug("calc period called")
        # 1. Mean shift Clustering  todo : Implement Spark DataFrame-native mean-shift clustering
        diff_array = np.array(diff_df.select(diff_col_name).collect())
        bandwidth = estimate_bandwidth(
            diff_array, quantile=0.5
        )  # todo : adjust quantile index
        bandwidth = round(bandwidth, 3)
        meanshift = MeanShift(bandwidth=bandwidth if bandwidth > 0 else None)
        cluster_labels = meanshift.fit_predict(diff_array)
        cluster_label_df = diff_df.sql_ctx.createDataFrame(
            enumerate(cluster_labels.tolist(), start=1), ["id", "label"]
        )
        clustered_diff_df = diff_df.withColumn(
            "id", F.row_number().over(Window.orderBy(F.monotonically_increasing_id()))
        ).join(
            cluster_label_df, on="id", how="inner"
        )  # todo: no partition window operation will move all data to a single partition
        clustered_diff_df.cache()

        # 2. approximated gcd  todo : enhance algorithm
        period = -1
        preset_unit = 1000  # todo : parameterize
        stopping_criteria: bool = False
        while not stopping_criteria:
            max_label = (
                clustered_diff_df.groupBy("label")
                .agg(F.count("*").alias("label_cnt"))
                .sort(F.col("label_cnt").desc())
                .first()
                .asDict()["label"]
            )
            period = (
                clustered_diff_df.filter(cluster_label_df["label"] == max_label)
                .agg(F.mean(diff_col_name).alias("period"))
                .first()
                .asDict()["period"]
                // preset_unit
                * preset_unit
            )
            stopping_criteria = True

        # 2-1. (optional) leave outlier cluster out # todo

        # 3. make a decision & report  # todo: enhance todo: report builder
        report = AnalysisReport()
        # 3-1. Period, Error, Regularity
        if period is not -1:
            report.parameters["period"] = period
            report.parameters["periodic_error"] = None
            report.parameters["regularity"] = "regular"
        else:
            report.parameters["regularity"] = "irregular"
        # 3-2. Report diff distribution, plot histogram
        report.parameters["distribution"] = clustered_diff_df
        # 3-3. Plot scatter one-dimension clustered time diff
        report_df = clustered_diff_df.withColumn("title", F.lit(diff_col_name))
        report.plots["cluster"] = px.scatter(
            report_df.toPandas(),
            y="title",
            x=diff_col_name,
            color="label",
            symbol="label",
        )

        # 3-4. Count missing value
        # 3-5. Outlier detect, plot, and do processing
        # 3-6. Explainable Validation report(p-value, Prediction interval...)
        logger.debug("calc period finish")
        return report



"""

"""

from pyspark.sql import DataFrame
from univariate.analyzer import AnalysisReport
from univariate.anomaly.detector import AnomalyDetector
from typing import Optional
import pyspark.sql.functions as F
import plotly.graph_objects as go
from plotly.graph_objects import Figure
from numpy import ndarray


class IQROutlierDetector(AnomalyDetector):
    """

    """
    @staticmethod
    def detect(ts: DataFrame, data_col_name: str, q1: Optional[float], median: Optional[float], q3: Optional[float], ad_hoc: bool = False) -> AnalysisReport:
        """

        :param ts:
        :param data_col_name:
        :param q1:
        :param median:
        :param q3:
        :param ad_hoc: If ad_hoc is True, than Q1, median, Q3 will be estminated on ts by itself and q1, median, q3 parameter will be ignored. Default is False.
        :return:
        """
        # validate parameters #1: if ad_hoc is false, then q1, median, q3 must be passed
        if not ad_hoc and (q1 is None or median is None or q3 is None):
            raise TypeError("If ad_hoc is false(default), then q1, median, q3 must be passed")

        # ad-hoc statistics Q1, median, Q3
        if ad_hoc:
            stat_dict = ts.agg(
                F.expr(f"percentile_approx({data_col_name}, 0.5)").alias(
                    "median"
                ),
                F.expr(f"percentile_approx({data_col_name}, 0.25)").alias(
                    "Q1"
                ),
                F.expr(f"percentile_approx({data_col_name}, 0.75)").alias(
                    "Q3"
                )
            ).first().asDict()
            # shadow parameters
            q1 = stat_dict['Q1']
            median = stat_dict['median']
            q3 = stat_dict['Q3']

        # detect outliers (classify)
        iqr = q3 - q1
        lower_fence = q1 - 1.5 * iqr
        upper_fence = q3 + 1.5 * iqr
        outlier_detected_df = ts.withColumn("is_anomaly_by_iqr_method", F.when(((F.col(data_col_name) < lower_fence) | (F.col(data_col_name) > upper_fence)), True)
                                            .otherwise(False))  # checked 221107 youngmin-an null -> false, # todo: testcase..

        # boxplot
        x = ts.select(data_col_name).toPandas().values.reshape((-1,))
        boxplot = IQROutlierDetector.__build_boxplot(x, q1, median, q3, lower_fence, upper_fence)

        # build report
        report = AnalysisReport()
        report.parameters["ground_truth_ts"] = ts
        report.parameters["outliers"] = outlier_detected_df
        report.plots["boxplot"] = boxplot

        return report

    @staticmethod
    def __build_boxplot(x: ndarray, q1: float, median: float, q3: float, lower_fence: float, upper_fence: float) -> Figure:
        fig = go.Figure()
        # Use x instead of y argument for horizontal plot
        fig.add_trace(go.Box(x=[x], y=["data"], boxpoints="outliers"))  # todo: feature name to y

        fig.update_traces(q1=[q1], median=[median],
                          q3=[q3], lowerfence=[lower_fence],
                          upperfence=[upper_fence], orientation="h")
        fig.update_layout(title_text="IQR method of outlier detection")

        fig.add_vline(
            x=q1, annotation_position="bottom", annotation_text="Q1", line_width=1, line_dash="dash"
        )
        fig.add_vline(
            x=median, annotation_position="top", annotation_text="median", line_width=1, line_dash="dash"
        )
        fig.add_vline(
            x=q3, annotation_position="bottom", annotation_text="Q3", line_width=1, line_dash="dash"
        )
        fig.add_vline(
            x=lower_fence, annotation_position="left top", annotation_text="lower_fence", line_width=1, line_dash="dash"
        )
        fig.add_vline(
            x=upper_fence, annotation_position="right top", annotation_text="upper_fence", line_width=1,
            line_dash="dash"
        )
        return fig



"""

"""

from univariate.sampling.strategy import SamplingStrategy
from pyspark.sql import DataFrame
from typing import Optional, List
from univariate.sampling.utils import find_start_timestamp, construct_range_df
from scipy import signal


class FourierMethodSampling(SamplingStrategy):
    """
    Sampling strategy using fourier method
    """
    @staticmethod
    def sample(ts: DataFrame, time_col_name: str, data_col_name: str, period: int, sampling_period: int, by_freq: bool, count_flag_cols: Optional[List[str]] = None, agg_point: Optional[str] = "middle", display_point: Optional[str] = "first") -> DataFrame:
        """
        Sampling method using fourier method implemented by scipy
        Note: It is not Spark-native
        :param ts:
        :param time_col_name:
        :param data_col_name:
        :param period:
        :param sampling_period:
        :param by_freq:
        :param count_flag_cols:
        :param agg_point:
        :param display_point:
        :return:
        """
        # first range's start timestamp
        first_timestamp = ts.sort(time_col_name).first().asDict()[time_col_name]
        first_start = find_start_timestamp(first_timestamp, by_freq, sampling_period)

        # last range's start timestamp
        last_timestamp = ts.sort(time_col_name).tail(1)[0].asDict()[time_col_name]
        last_start = find_start_timestamp(last_timestamp, by_freq, sampling_period)

        # fit
        ts_pd = ts.sort(time_col_name).select(time_col_name, data_col_name).toPandas()
        # range df
        range_df = construct_range_df(sampling_period, first_start, last_start)

        # new point

        #
        # result


"""

"""

from univariate.strategy.agreement import AgreementStrategy
from univariate.analyzer import AnalysisReport
from pyspark.sql import DataFrame


class BlandAltmanPlot(AgreementStrategy):
    """ """

    @classmethod
    def measure_agreement(cls, sample: DataFrame) -> AnalysisReport:
        pass



"""

"""

from abc import ABCMeta, abstractmethod
from enum import Enum
from pyspark.sql import DataFrame
from typing import Optional, List


class SamplingStrategyType(Enum):
    mean_downsampling = "MeanDownsampling"
    linear_upsampling = "LinearUpsampling"
    # fourier_method_sampling = "FourierMethodSampling"


class SamplingStrategy(metaclass=ABCMeta):
    """

    """
    @staticmethod
    @abstractmethod
    def sample(ts: DataFrame, time_col_name: str, data_col_name: str, period: int, sampling_period: int, by_freq: bool, count_flag_cols: List[str], **kwargs) -> DataFrame:
        """

        :param ts:
        :param time_col_name:
        :param data_col_name:
        :param period:
        :param sampling_period:
        :param by_freq:
        :param count_flag_cols:
        :return:
        """
        pass



from univariate.strategy.period.period_calc_strategy import PeriodCalcType
from univariate.strategy.period.period_calc_strategy import PeriodCalcStrategy
from univariate.strategy.period.clustering_and_gcd import ClusteringAndApproximateGCD



"""

"""

from univariate.missing_value.handler.strategy import MissingValueHandleStrategy, MissingValueHandleStrategyType
from pyspark.sql import DataFrame
import pyspark.sql.functions as F
from univariate.analyzer import AnalysisReport
from univariate.analyzer import DescriptiveStatAnalyzer
from typing import Optional


class ModeImputation(MissingValueHandleStrategy):
    """ """

    def handle(
        self, missed_ts: DataFrame, time_col_name: str, data_col_name, handled_col_name: str, mode: Optional[float] = None, ad_hoc: bool=False
    ) -> DataFrame:
        """

        :param missed_ts:
        :param time_col_name:
        :param data_col_name:
        :param handled_col_name:
        :param mode: It can only be None when ad_hoc is False. If ad_hoc is False and mode is None, then throw exception
        :param ad_hoc:
        :return:
        """
        # validate: if non-ad-hoc (ad-hoc=False) run and mode is None then it is invalid request
        if mode is None and not ad_hoc:
            # todo: error log to spark logger
            raise TypeError("When Imputation is not ad-hoc, mode value must be passed by argument")
        if ad_hoc:
            descriptive_stat_report = DescriptiveStatAnalyzer().analyze(missed_ts, time_col_name=time_col_name, data_col_name=data_col_name)  # todo: selective field
            mode = descriptive_stat_report.parameters["mode"]
        filled_ts = missed_ts.withColumn(handled_col_name, F.coalesce(data_col_name, F.lit(round(mode, 3))))
        return filled_ts



"""

"""


from pyspark.sql import DataFrame
from univariate.sampling.strategy import SamplingStrategy, MeanDownsampling, LinearUpsampling
from univariate.sampling.utils import freq_to_period_map
from typing import List, Optional




class Sampler:
    """

    """

    @staticmethod
    def sample_by_freq_alias(ts: DataFrame, time_col_name: str, data_col_name: str, period: int, freq: str, count_flag_cols: Optional[List[str]] = []) -> DataFrame:
        """

        :return:
        """
        assert Sampler.__validate_freq_alias(freq)
        sampling_strategy = Sampler.__instantiate_sampling_strategy(period, freq_to_period_map[freq])
        return sampling_strategy.sample(ts, time_col_name, data_col_name, period, freq_to_period_map[freq], True, count_flag_cols)

    @staticmethod
    def __validate_freq_alias(freq: str) -> bool:
        """

        :param freq: frequency alias
        :return: if freq is valid return true, else exception will be raised

        Note: please see https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
        (221117) Support alias: 'A', 'Q', 'M', 'W', 'D', 'H'
        """
        # todo: if you want more aliases issue a ticket
        supports = ['A', 'Q', 'M', 'W', 'D', 'H']
        if freq not in supports:
            raise ValueError(f"freq {freq} is not supported now, supports {supports} (if you want more aliases issue a ticket)")

        return True

    @staticmethod
    def __instantiate_sampling_strategy(period: int, sampling_period) -> SamplingStrategy:
        """

        :param period:
        :param sampling_period:
        :return:
        """
        # todo: parameterize or auto-decision instantiation
        if period <= sampling_period:
            strategy = MeanDownsampling()
        else:
            strategy = LinearUpsampling()

        return strategy



from univariate.missing_value.strategy.unmarked_detection.unmarked_missing_value_detection_strategy import (
    UnmarkedMissingValueDetectionStrategyType,
)
from univariate.missing_value.strategy.unmarked_detection.unmarked_missing_value_detection_strategy import (
    UnmarkedMissingValueDetectionStrategy,
)
from univariate.missing_value.strategy.unmarked_detection.divide_and_round import (
    DivideAndRound,
)
from univariate.missing_value.strategy.unmarked_detection.do_not_detect import (
    DoNotDetect,
)



from univariate.missing_value.handler.strategy.missing_value_handle_strategy import MissingValueHandleStrategyType, MissingValueHandleStrategy
from univariate.missing_value.handler.strategy.mean_imputation import MeanImputation
from univariate.missing_value.handler.strategy.median_imputation import MedianImputation
from univariate.missing_value.handler.strategy.mode_imputation import ModeImputation
from univariate.missing_value.handler.strategy.locf_imputation import LOCFImputation
from univariate.missing_value.handler.strategy.bocf_imputation import BOCFImputation
from univariate.missing_value.handler.strategy.linear_interploation import LinearInterpolation
from univariate.missing_value.handler.strategy.lagrange_interpolation import LagrangeInterpolation
from univariate.missing_value.handler.strategy.cubic_spline_interpolation import CubicSplineInterpolation
from univariate.missing_value.handler.strategy.deletion import Deletion



"""

"""
from pyspark.sql import DataFrame
import pyspark.sql.functions as F
from pyspark.sql.functions import udf
from pyspark.sql.types import ArrayType, LongType
from pyspark.sql.window import Window
import numpy as np
from univariate.missing_value.strategy.unmarked_detection import (
    UnmarkedMissingValueDetectionStrategy,
)


@udf(ArrayType(LongType()))
def span_udf(start, end, num):
    return (
        np.round(np.linspace(start=start, stop=end, num=int(num), endpoint=False)[1:])
        .astype(np.int_)
        .tolist()
    )


class DivideAndRound(UnmarkedMissingValueDetectionStrategy):
    """ """

    @staticmethod
    def detect(
        ts: DataFrame,
        time_col_name: str,
        data_col_name: str,
        period: int
    ) -> DataFrame:
        """

        :param ts:
        :param time_col_name:
        :param data_col_name:
        :param period:
        :return: 'Null' interleaved dataframe as detected unmarked missing values
        """
        # todo: enhance algorithm
        # 1. Divide and Round
        bias = -0.2
        divide_df = (
            ts.select([time_col_name, data_col_name])
            .sort(time_col_name)
            .withColumn(
                "lagged_time", F.lag(time_col_name).over(Window.orderBy(time_col_name))
            )
            .withColumn(
                "ratio_by_period",
                (F.col(time_col_name) - F.col("lagged_time"))
                / period,
            )
        )
        round_df = divide_df.filter("ratio_by_period >= 1.7").withColumn(
            "estimated_missing_count", F.round(F.col("ratio_by_period") - F.lit(bias))
        )

        # 2. Interleave null values
        estimated_unmarked_missing_df = round_df.withColumn(
            "missing_timestamps",
            span_udf("lagged_time", time_col_name, "estimated_missing_count"),
        ).select(
            F.explode("missing_timestamps").alias(time_col_name),
            F.lit(None).alias(data_col_name),
        )

        # 3. merge and sort
        interleaved_df = (
            ts.select([data_col_name, time_col_name])
            .unionByName(estimated_unmarked_missing_df)
            .sort(time_col_name)
        )
        return interleaved_df



"""

"""

from enum import Enum
from abc import ABCMeta, abstractmethod
from pyspark.sql import DataFrame
from univariate.analyzer import AnalysisReport


class MissingValueHandleStrategyType(Enum):
    deletion = "Deletion"
    mean_imputation = "MeanImputation"
    median_imputation = "MedianImputation"
    mode_imputation = "ModeImputation"
    locf_imputation = "LOCFImputation"
    bocf_imputation = "BOCFImputation"
    # constant_imputation = "ConstantImputation"
    # seasonality-based imputation

    linear_interpolation = "LinearInterpolation"
    lagrange_interpolation = "LagrangeInterpolation"
    cubic_spline_interpolation = "CubicSplineInterpolation"

    # k_nn_imputation = "KNNImputation" #multi
    # mice_imputation = "MICEImputation" #multi
    # regression # multi
    # stochastic regressioin # multi
    # hot_deck_imputation = "HOT_DECK_IMPUTATION" multi

    # todo: continous update
    # todo: where cross-informated strategy will be
    # todo: enhance
    # todo: process exceptions with sufficient test coverage


class MissingValueHandleStrategy(metaclass=ABCMeta):
    """
    Missing
    """

    @abstractmethod
    def handle(
        self, missed_ts: DataFrame, time_col_name: str, data_col_name: str, handled_col_name: str, **kwargs
    ) -> DataFrame:
        pass



"""

"""

from abc import ABCMeta
from pyspark.sql import DataFrame
from univariate.analyzer import AnalysisReport
from typing import Optional

class AnomalyDetector(metaclass=ABCMeta):
    """

    """
    @staticmethod
    def detect(ts: DataFrame, data_col_name: str, q1: Optional[float], median: Optional[float], q3: Optional[float], ad_hoc: bool=False) -> AnalysisReport:
        """

        :param ts:
        :param data_col_name:
        :param q1:
        :param median:
        :param q3:
        :param ad_hoc:
        :return:
        """
        pass



"""

"""

from univariate.duplicate import DuplicateReport
from pyspark.sql import DataFrame
import logging

logger = logging.getLogger()


class DuplicateProcessor:
    """

    """

    @staticmethod
    def detect_duplicates(ts: DataFrame, time_col_name: str, data_col_name: str) -> DuplicateReport:
        """

        :param ts:
        :param time_col_name:
        :param data_col_name:
        :return:
        """
        duplicate_df = ts.groupBy([time_col_name, data_col_name]).count().where('count > 1')
        report = DuplicateReport(duplicate_df)
        return report

    @staticmethod
    def drop_duplicates(ts: DataFrame) -> DataFrame:
        """

        :param ts:
        :return:
        """
        distinct_df = ts.distinct()
        logger.info("Duplicates dropped")
        return distinct_df



from univariate.analyzer.analysis_report import AnalysisReport
from univariate.analyzer.analyzer import Analyzer
from univariate.analyzer.regularity_analyzer import RegularityAnalyzer
from univariate.analyzer.descriptive_stat_analyzer import DescriptiveStatAnalyzer
from univariate.analyzer.autocorrelation_analyzer import AutocorrelationAnalyzer
from univariate.analyzer.decomposition_analyzer import DecompositionAnalyzer



"""

"""

from abc import ABCMeta, abstractmethod
from enum import Enum
from univariate.analyzer import AnalysisReport
from pyspark.sql import DataFrame


class AgreementType(Enum):
    BLAND_ALTMAN = "BlandAltmanPlot"


class AgreementStrategy(metaclass=ABCMeta):
    """
    Abstract strategy for making a decision that sample set values are agreed.
    """

    @classmethod
    def measure_agreement(cls, sample: DataFrame) -> AnalysisReport:
        pass



"""
    Univariate Time Series Analyzer module
"""

from pyspark.sql import DataFrame
from typing import Optional, List
from univariate.hook import Hook
from univariate.analyzer import (
    AnalysisReport,
    Analyzer,
    RegularityAnalyzer,
    DescriptiveStatAnalyzer,
)
from univariate.ts_validator import Validator
from univariate.duplicate import DuplicateProcessor, DuplicateReport
import logging

logger = logging.getLogger()


class UnivariateAnalyzer:
    """
    Univariate Time Series Analyzer
    """

    def __init__(
        self,
        ts: DataFrame,
        val_col: str = "value",
        time_col: str = "time",
        hooks: Optional[List[Hook]] = [],
        validators: Optional[List[Validator]] = [],
        drop_duplicates: bool = True,
    ):
        """

        :param ts: Spark dataframe that has one datatime column, and has one timeseries value(float or number) column
        :param val_col: string column name containing timeseries values
        :param time_col: string column name containing pyspark.sql.DatetimeType
        :param hooks: list of hooks that want to be notified analysis reports
        """
        logger.debug(
            f"Setup Analyzer start, ts: {ts}, val_col: {val_col}, time_col: {time_col}, hooks: {hooks or '[]'}"
        )
        self.validators: List[
            Validator
        ] = validators  # todo : default validator  # todo : who is creator that has responsibility
        self.duplicate_report: DuplicateReport = DuplicateProcessor.detect_duplicates(ts=ts, time_col_name=time_col, data_col_name=val_col)
        if drop_duplicates and self.duplicate_report.duplicated:
            ts = DuplicateProcessor.drop_duplicates(ts=ts)
        self.__validate_ts(ts=ts, time_col_name=time_col, data_col_name=val_col)
        self.hooks: List[Hook] = hooks  # todo : default post analysis hook
        self.ts: DataFrame = ts.sort(time_col).select([time_col, val_col])
        self.analysis_queue: List[Analyzer] = list()
        self.time_col_name: str = time_col
        self.data_col_name: str = val_col

        regularity_analyzer: Analyzer = RegularityAnalyzer()
        self.regularity_report: AnalysisReport = regularity_analyzer.analyze(
            ts=self.ts, time_col_name=time_col
        )
        self.__notify_report(self.regularity_report)
        self.__enqueue_analysis_job()
        logger.debug("Setup Analyzer finish")

    def __validate_ts(self, ts: DataFrame, time_col_name: str, data_col_name: str):
        """
        If has validation error, it raise exception
        :param val_col:
        :param time_col:
        :return:
        """
        logger.debug("Validate ts")
        for validator in self.validators:
            validator(ts=ts, time_col_name=time_col_name, data_col_name=data_col_name)
        # map(lambda validator: validator(ts=ts, time_col_name=time_col_name, data_col_name=data_col_name), self.validators)

    def __notify_report(self, report: AnalysisReport):
        map(lambda hook: hook.do_post_analysis(report), self.hooks)
        logger.debug("Notify all hook")
        logger.debug("Notify hook finish")

    def __enqueue_analysis_job(self):
        """

        :return:
        """
        logger.debug("enqueue analysis job start")
        if self.regularity_report.parameters["regularity"] == "regular":
            logger.debug("regularity: regular")
            self.analysis_queue.append(DescriptiveStatAnalyzer())
            pass
        elif self.regularity_report.parameters["regularity"] == "irregular":
            logger.debug("regularity: irregular")
            pass
        logger.debug("enqueue analysis job finish")

    def analyze(self):
        """

        :return:
        """
        logger.debug("analyze called")

        reports = list(
            map(
                lambda job: job.analyze(
                    ts=self.ts,
                    time_col_name=self.time_col_name,
                    data_col_name=self.data_col_name,
                ),
                self.analysis_queue,
            )
        )  # todo : Explore parallezing with spark sub(child) context & python asyncio?
        map(
            lambda report: map(lambda hook: hook.do_post_analysis(report), self.hooks),
            reports,
        )
        """
        reports = []
        for job in self.analysis_queue:
            report = job.analyze(ts=self.ts, time_col_name=self.time_col_name, data_col_name=self.data_col_name)
            reports.append(report)
        """  # todo : clean dummy code
        logger.debug("analyze return")
        return reports



from setuptools import setup, find_packages


__version__ = "0.14.0"

setup(
    name="aicns-univariate-analyzer",
    version=__version__,
    description="Univariate time series analyzer library package in AICNS project",
    author="Youngmin An",
    author_email="youngmin.develop@gmail.com",
    license="Apache License 2.0",
    packages=find_packages(),
    install_requires=[
        "pandas==1.1.5",
        "plotly==5.10.0",
        "pyspark>=3.1.0",
        "scikit-learn==0.24.2",
        "statsmodels==0.13.2",
        "pendulum==2.1.2"
    ],
)



"""

"""

from abc import ABCMeta, abstractmethod
from enum import Enum
from univariate.analyzer import AnalysisReport
from pyspark.sql import DataFrame


class PeriodCalcType(Enum):
    ClusteringAndApproximateGCD = "ClusteringAndApproximateGCD"


class PeriodCalcStrategy(metaclass=ABCMeta):
    """
    Abstract strategy for calculating regular period between each observation if time series is regular
    """

    @classmethod
    def calc_period(cls, diff_df: DataFrame, diff_col_name: str) -> AnalysisReport:
        pass



"""

"""

from univariate.analyzer import Analyzer
from univariate.analyzer import AnalysisReport
from univariate.strategy.period import PeriodCalcStrategy, PeriodCalcType
from typing import Dict, Optional
from enum import Enum
import importlib
from pyspark.sql import DataFrame
from pyspark.sql.window import Window
import pyspark.sql.functions as F
import logging

logger = logging.getLogger()


class RegularityAnalyzer(Analyzer):
    """ """

    def __init__(
        self,
        period_strategy_type: PeriodCalcType = PeriodCalcType.ClusteringAndApproximateGCD,
    ):
        """

        :param period_strategy_type: Enum val for calc period strategy, default ClusteringAndApproximateGCD
        """
        concrete_cls = getattr(
            importlib.import_module("univariate.strategy.period"),
            period_strategy_type.value,
        )
        self.period_strategy: PeriodCalcStrategy = concrete_cls()

    def analyze(self, ts: DataFrame, **kwargs) -> AnalysisReport:
        """

        :param ts: Time series spark dataframe containing timestamp, data
        :param time_col_name:
        :return:
        """
        logger.debug("analyze called")

        # [AINCS-72] No decision when few data
        # todo: where from threshold config
        threshold = 5
        if ts.count() < threshold:
            report = AnalysisReport()
            report.parameters["period"] = None
            report.parameters["periodic_error"] = None
            report.parameters["regularity"] = "no_decision"
            return report

        diff_col_name = kwargs["time_col_name"] + "_diff"
        differenced_timestamp = self.__make_difference_series_of_timestamp(
            ts, kwargs["time_col_name"], diff_col_name
        )
        logger.debug("analyze finish")
        return self.period_strategy.calc_period(differenced_timestamp, diff_col_name)

    def __make_difference_series_of_timestamp(
        self, ts: DataFrame, time_col_name: str, diff_col_name: str
    ) -> DataFrame:
        """

        :param ts:
        :param time_col_name:
        :param diff_col_name:
        :return:
        """
        logger.debug("make difference series of timestamp called")
        window = Window.partitionBy().orderBy(time_col_name)

        diff_df = (
            ts.withColumn(
                diff_col_name,
                F.col(time_col_name) - F.lag(F.col(time_col_name), 1).over(window),
            )
            .select(diff_col_name)
            .na.drop()
        )
        logger.debug("make difference series of timestamp finish")
        return diff_df
        # todo : think less column exception



"""

"""

from univariate.analyzer import Analyzer
from univariate.analyzer import AnalysisReport
from pyspark.sql import DataFrame
from statsmodels.tsa.stattools import pacf, acf
import plotly.graph_objects as go
import numpy as np


class AutocorrelationAnalyzer(Analyzer):
    """ """
    def analyze(self, ts: DataFrame, data_col_name: str, regularity_report: AnalysisReport) -> AnalysisReport:
        """
        Analyze autocorrelation. Missing values will be ignored
        :param ts: DataFrame has to be sorted by col(time_col_name)
        :param time_col_name:
        :param data_col_name:
        :param regularity_report: Yielded by Univariate Analyzer
        :return:
        """
        # todo: Deep-unstand how statsmodels handle missing values
        # todo: this version is only for regular, so have to diverge strategies by regular and irregular

        num_lags = ts.count()  # todo: adjusted nlags (priority: high)
        plot_pacf = False  # todo: what is pacf, what is diff
        corr_array = pacf(ts.select(data_col_name).toPandas().values.reshape((-1,)), alpha=0.05, missing="conservative", nlags=num_lags) if plot_pacf else acf(ts.select(data_col_name).toPandas().values.reshape((-1,)),
                                                                                                       alpha=0.05,
                                                                                                       missing="conservative",
                                                                                                       nlags=num_lags)
        lower_confident_interval = corr_array[1][:, 0] - corr_array[0]  # todo: have to know which algorithm used to find the confidence interval
        upper_confident_interval = corr_array[1][:, 1] - corr_array[0]

        correlogram = go.Figure()
        [correlogram.add_scatter(x=(x, x), y=(0, corr_array[0][x]), mode='lines', line_color='#3f3f3f', line_width=0.6)
         for x in range(len(corr_array[0]))]  # todo: adjusted line_width
        correlogram.add_scatter(x=np.arange(len(corr_array[0])), y=corr_array[0], mode='markers', marker_color='#1f77b4')
        correlogram.add_scatter(x=np.arange(len(corr_array[0])), y=upper_confident_interval, mode='lines', line_color='rgba(255,255,255,0)')
        correlogram.add_scatter(x=np.arange(len(corr_array[0])), y=lower_confident_interval, mode='lines', fillcolor='rgba(32, 146, 230,0.3)',
                        fill='tonexty', line_color='rgba(255,255,255,0)')
        correlogram.update_traces(showlegend=False)
        # fig.update_xaxes(range=[-1,42])
        correlogram.update_yaxes(zerolinecolor='#000000')

        title = 'Partial Autocorrelation (PACF)' if plot_pacf else 'Autocorrelation (ACF)'
        correlogram.update_layout(title=title)

        # Build report
        report = AnalysisReport()

        report.parameters["acf"] = corr_array[0].tolist()
        report.parameters["num_lags"] = num_lags
        report.parameters["lower_confident_interval"] = lower_confident_interval.tolist()
        report.parameters["upper_confident_interval"] = upper_confident_interval.tolist()

        report.plots["correlogram"] = correlogram

        return report
